@@ -1,12 +1,12 @@
 """
-Open Edison Server
-
-Simple FastAPI + FastMCP server for single-user MCP proxy.
-No multi-user support, no complex routing - just a straightforward proxy.
+Open Edison MCP Proxy Server
+Main server entrypoint for FastAPI and FastMCP integration.
+See README for usage and configuration details.
 """
 
 import asyncio
 import json
+import signal
 import traceback
 from collections.abc import Awaitable, Callable, Coroutine
 from contextlib import suppress
@@ -25,18 +25,24 @@ from fastapi.responses import (
 )
 from fastapi.security import HTTPAuthorizationCredentials, HTTPBearer
 from fastapi.staticfiles import StaticFiles
+from fastmcp import Client as FastMCPClient
 from fastmcp import FastMCP
 from loguru import logger as log
 from pydantic import BaseModel, Field
 
 from src import events
-from src.config import Config, MCPServerConfig, clear_json_file_cache
+from src.config import Config, MCPServerConfig, clear_json_file_cache, get_config_json_path
 from src.config import get_config_dir as _get_cfg_dir  # type: ignore[attr-defined]
+from src.mcp_stdio_capture import (
+    install_stdio_client_stderr_capture as _install_stdio_capture,
+)
 from src.middleware.session_tracking import (
     MCPSessionModel,
     create_db_session,
 )
 from src.oauth_manager import OAuthStatus, get_oauth_manager
+from src.oauth_override import OpenEdisonOAuth
+from src.permissions import Permissions
 from src.single_user_mcp import SingleUserMCP
 from src.telemetry import initialize_telemetry, set_servers_installed
 
@@ -45,6 +51,9 @@ _security = HTTPBearer()
 _auth_dependency = Depends(_security)
 
 
+_install_stdio_capture()
+
+
 class OpenEdisonProxy:
     """
     Open Edison Single-User MCP Proxy Server
@@ -276,6 +285,12 @@ class OpenEdisonProxy:
                 # Clear cache for the config file, if it was config.json
                 if name == "config.json":
                     clear_json_file_cache()
+                elif name in (
+                    "tool_permissions.json",
+                    "resource_permissions.json",
+                    "prompt_permissions.json",
+                ):
+                    Permissions.clear_permissions_file_cache()
 
                 return {"status": "ok"}
             except Exception as e:  # noqa: BLE001
@@ -363,6 +378,9 @@ class OpenEdisonProxy:
         log.info(f"FastAPI management API on {self.host}:{self.port + 1}")
         log.info(f"FastMCP protocol server on {self.host}:{self.port}")
 
+        # Print location of config
+        log.info(f"Config file location: {get_config_json_path()}")
+
         initialize_telemetry()
 
         # Ensure the sessions database exists and has the required schema
@@ -397,6 +415,7 @@ class OpenEdisonProxy:
             host=self.host,
             port=self.port + 1,
             log_level=Config().logging.level.lower(),
+            timeout_graceful_shutdown=0,
         )
         fastapi_server = uvicorn.Server(fastapi_config)
         servers_to_run.append(fastapi_server.serve())
@@ -408,13 +427,27 @@ class OpenEdisonProxy:
             host=self.host,
             port=self.port,
             log_level=Config().logging.level.lower(),
+            timeout_graceful_shutdown=0,
         )
         fastmcp_server = uvicorn.Server(fastmcp_config)
         servers_to_run.append(fastmcp_server.serve())
 
         # Run both servers concurrently
         log.info("🚀 Starting both FastAPI and FastMCP servers...")
-        _ = await asyncio.gather(*servers_to_run)
+        loop = asyncio.get_running_loop()
+
+        def _trigger_shutdown(signame: str) -> None:
+            log.info(f"Received {signame}. Forcing shutdown of all servers...")
+            for srv in (fastapi_server, fastmcp_server):
+                with suppress(Exception):
+                    srv.force_exit = True  # type: ignore[attr-defined] # noqa
+                    srv.should_exit = True  # type: ignore[attr-defined] # noqa
+
+        for sig in (signal.SIGINT, signal.SIGTERM):
+            with suppress(Exception):
+                loop.add_signal_handler(sig, _trigger_shutdown, sig.name)
+
+        await asyncio.gather(*servers_to_run, return_exceptions=False)
 
     def _register_routes(self, app: FastAPI) -> None:
         """Register all routes for the FastAPI app"""
@@ -519,11 +552,8 @@ class OpenEdisonProxy:
         warms the lists to ensure subsequent list calls reflect current state.
         """
         try:
-            mcp = self.single_user_mcp
-            # Warm managers so any internal caches are refreshed
-            await mcp._tool_manager.list_tools()  # type: ignore[attr-defined]
-            await mcp._resource_manager.list_resources()  # type: ignore[attr-defined]
-            await mcp._prompt_manager.list_prompts()  # type: ignore[attr-defined]
+            clear_json_file_cache()
+            Permissions.clear_permissions_file_cache()
             return {"status": "ok"}
         except Exception as e:  # noqa: BLE001
             log.error(f"Failed to process permissions-changed: {e}")
@@ -638,14 +668,29 @@ class OpenEdisonProxy:
                 )
 
                 sessions: list[dict[str, Any]] = []
+                has_warned_about_missing_created_at = False
                 for row_model in results:
                     row = cast(Any, row_model)
                     tool_calls_val = row.tool_calls
                     data_access_summary_val = row.data_access_summary
+                    created_at_val = None
+                    if isinstance(data_access_summary_val, dict):
+                        created_at_val = data_access_summary_val.get("created_at")  # type: ignore[assignment]
+                    if (
+                        created_at_val is None
+                        and isinstance(tool_calls_val, list)
+                        and tool_calls_val
+                        and not has_warned_about_missing_created_at
+                    ):
+                        has_warned_about_missing_created_at = True
+                        log.warning(
+                            "created_at is missing, will have sessions with unknown timestamps"
+                        )
                     sessions.append(
                         {
                             "session_id": row.session_id,
                             "correlation_id": row.correlation_id,
+                            "created_at": created_at_val,
                             "tool_calls": tool_calls_val
                             if isinstance(tool_calls_val, list)
                             else [],
@@ -769,9 +814,15 @@ class OpenEdisonProxy:
             raise AttributeError(f"Proxy does not expose list method for {kind}")
 
         async def list_all() -> tuple[list[Any], list[Any], list[Any]]:
-            tools = await _call_list("tools")
-            resources = await _call_list("resources")
-            prompts = await _call_list("prompts")
+            tools, resources, prompts = await asyncio.gather(
+                _call_list("tools"),
+                _call_list("resources"),
+                _call_list("prompts"),
+                return_exceptions=True,
+            )
+            tools = tools if isinstance(tools, list) else []
+            resources = resources if isinstance(resources, list) else []
+            prompts = prompts if isinstance(prompts, list) else []
             return tools, resources, prompts
 
         timeout = body.timeout_s if isinstance(body.timeout_s, (int | float)) else 20.0
@@ -953,12 +1004,8 @@ class OpenEdisonProxy:
 
             log.info(f"🔗 Testing connection to {server_name} at {remote_url}")
 
-            # Import FastMCP client for testing
-            from fastmcp import Client as FastMCPClient
-            from fastmcp.client.auth import OAuth
-
             # Create OAuth auth object
-            oauth = OAuth(
+            oauth = OpenEdisonOAuth(
                 mcp_url=remote_url,
                 scopes=scopes,
                 client_name=client_name or "OpenEdison MCP Gateway",
